function [R, T, K, c, a, tanftr,U,x] = CalibrateHyperbolicMirrorCamera(cp, truepoints, image_size,c,a, param)
%Joshua Stough
%COMP290-89, Pollefeys
%April 13, 2003
%This program finds the camera calibration matrices for the 360 deg 
%hyperbolic (at least that's the theory) mirror camera.  The program
%is provided the 3D world coordinates whose image is cp (the calibration
%points selected in PickCalibrationPoints for example).  It assumes no
%error in the points, finding the best rotation R, translation T, 
%intrinsic camera matrix K, and the hyperbola parameters c and a for
%the provided correspondences.

%To project from the world to the image coordinates, R*X + T.  Then do
%the non-linear projection that is the hyperbolic mirror, which gives
%orthogonal pixel coords for the projected points.  Then K times that
%gives the final coords (which should then be close to those in cp).

global x_world x_image

if (sum(size(cp) ~= size(truepoints)) - 1 ~= 0)
    error('ERROR: the calibration and true points arrays are not the same size.');
end

if (nargin < 6), param = 6500; end
if (nargin < 5), a = 4.1; end
if (nargin < 4), c = 4.15; end

s = image_size;

%As per the paper, intialize the pipeline from world to image, then 
%optimize.

parameters = zeros(12,1);
%The 3 params of rotation, 3 of translation, 2 for the image center,
%1 scale ratio of the image axes, and 2 hyperbola parameters and the 
%parameter for the ratio of orthogonal distance to the optical axis in
%.5" units verses pixels (that ratio).

%Initialize
%First, rotation parameters
parameters(1:3) = [0 0 -pi/24]';
%Translation
parameters(4:6) = [0 0 0]';

%image center and axes scale
parameters(7:9) = [s(1)/2 s(2)/2 1]';

%the initial estimate of the a and c parameters.
%I found that the radius of the mirror (3 half-inches)
%took 1050 pixels I think.  So that ratio says multiply
%a camera coordinate |(x,y)| by 350 to get a magnitude in
%pixels.
%Or this is used for stereographic projection, and I need
%the ratio of pixels to radians.  The angle .34 radians 
%corresponded to 1050 pixels, so 1050/tan(.34/2) = 6116.856
parameters(10:12) = [a c param]';


%%%%%%%%%%%%%%%%%%%%%%
%HERE ASSUMPTIONS MADE REGARDING THE SIZE OF THE DATA.
%%%%%%%%%%%%%%%%%%%%%%

%Expect truepoints to be 3x50; cp 2x50 (as read from file).
x_world = truepoints;
x_image = cp;

%if (size(x_world, 3) == 1)
%    x_world = reshape(x_world, 3,10,5);
%end

%if (size(x_image, 3) == 1)
%    x_image = x_image - [s(2)/2 s(1)/2]'*ones(1,50);
%    x_image = reshape(x_image, 2,10,5);
%else
%    for i = 1:5
%        x_image(:,:,i) = x_image(:,:,i) - [s(2)/2 s(1)/2]'*ones(1,10);
%    end
%end

%center the calibration points.  Maybe that's part of the K matrix.
%x_image = x_image - [s(1)/2 s(2)/2]'*ones(1,50);

%Now we're ready to start the non-lin optmization.

x0 = parameters;

x = lsqnonlin(@HyperbolicMirrorCost, x0);
%x = fminsearch(@HyperbolicMirrorCost, x0);


[d,U] = GetHyperbolicResidual(x);

disp(sprintf('Final total error: %f', sum(d.^2)));
disp(sprintf('average and median error: %f %f', mean(d), median(d)));

disp('Graphing the error per vertex.');
figure, plot(1:50, d);

%Here's something just added.  Now run the non lin again, on just the 
%the top 3/4 cases.
[c,i] = sort(d);

s = ceil((3/4)*size(i,1));

x_world = x_world(:, i(1:s));
x_image = x_image(:, i(1:s));

x2 = lsqnonlin(@HyperbolicMirrorCost, x0);
[d,U2] = GetHyperbolicResidual(x);
disp(sprintf('Error of inliers: %f', sum(d.^2)));
disp(sprintf('average and median error, inliers: %f %f', mean(d), median(d)));

disp('The two x"s');
[x x2]



%Build the R,T,K,c,a model params.

c = cos(x(1));
s = sin(x(1));
rotx = [1 0 0; 0 c -s; 0 s c];

c = cos(x(2));
s = sin(x(2));
roty = [c 0 s; 0 1 0; -s 0 c];

c = cos(x(3));
s = sin(x(3));
rotz = [c -s 0; s c 0; 0 0 1];

R = rotx*roty*rotz;

T = x(4:6);

a = x(10);
c = x(11);
tanftr = x(12);

Kinv = [1 0 -x(7); 0 x(9) -x(9)*x(8); 0 0 1];

K = inv(Kinv);
